from yee88.telegram import (
    TelegramCallbackQuery,
    TelegramIncomingMessage,
    parse_incoming_update,
)
from yee88.telegram.api_models import (
    CallbackQuery,
    CallbackQueryMessage,
    Chat,
    Document,
    Message,
    MessageReply,
    PhotoSize,
    Sticker,
    Update,
    User,
    Video,
    Voice,
)


def test_parse_incoming_update_maps_fields() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            text="hello",
            chat=Chat(id=123, type="supergroup", is_forum=True),
            from_=User(id=99),
            reply_to_message=MessageReply(
                message_id=5,
                text="prev",
                from_=User(id=77, is_bot=True, username="ReplyBot"),
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.transport == "telegram"
    assert msg.chat_id == 123
    assert msg.message_id == 10
    assert msg.text == "hello"
    assert msg.reply_to_message_id == 5
    assert msg.reply_to_text == "prev"
    assert msg.reply_to_is_bot is True
    assert msg.reply_to_username == "ReplyBot"
    assert msg.sender_id == 99
    assert msg.thread_id is None
    assert msg.is_topic_message is None
    assert msg.chat_type == "supergroup"
    assert msg.is_forum is True
    assert msg.voice is None
    assert msg.document is None
    assert msg.raw
    assert msg.raw["message_id"] == 10


def test_parse_incoming_update_ignores_implicit_topic_reply() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=187,
            message_thread_id=163,
            is_topic_message=True,
            text="Hello",
            chat=Chat(id=123, type="supergroup", is_forum=True),
            from_=User(id=99),
            reply_to_message=MessageReply(
                message_id=163,
                from_=User(id=77, is_bot=True, username="TakopiBot"),
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.thread_id == 163
    assert msg.is_topic_message is True
    assert msg.reply_to_message_id is None
    assert msg.reply_to_text is None
    assert msg.reply_to_is_bot is None
    assert msg.reply_to_username is None


def test_parse_incoming_update_filters_non_matching_chat() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            text="hello",
            chat=Chat(id=123, type="private"),
        ),
    )

    assert parse_incoming_update(update, chat_id=999) is None


def test_parse_incoming_update_filters_non_text_and_non_voice() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            chat=Chat(id=123, type="private"),
        ),
    )

    assert parse_incoming_update(update, chat_id=123) is None


def test_parse_incoming_update_voice_message() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            chat=Chat(id=123, type="private"),
            voice=Voice(
                file_id="voice-id",
                duration=3,
                mime_type="audio/ogg",
                file_size=1234,
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.text == ""
    assert msg.voice is not None
    assert msg.voice.file_id == "voice-id"
    assert msg.voice.mime_type == "audio/ogg"
    assert msg.voice.file_size == 1234
    assert msg.voice.duration == 3


def test_parse_incoming_update_document_message() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            caption="/file put incoming/doc.txt",
            chat=Chat(id=123, type="private"),
            document=Document(
                file_id="doc-id",
                file_name="doc.txt",
                mime_type="text/plain",
                file_size=4321,
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.text == "/file put incoming/doc.txt"
    assert msg.document is not None
    assert msg.document.file_id == "doc-id"
    assert msg.document.file_name == "doc.txt"
    assert msg.document.mime_type == "text/plain"
    assert msg.document.file_size == 4321


def test_parse_incoming_update_photo_message() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            caption="/file put incoming/photo.jpg",
            chat=Chat(id=123, type="private"),
            photo=[
                PhotoSize(
                    file_id="small",
                    file_size=100,
                    width=90,
                    height=90,
                ),
                PhotoSize(
                    file_id="large",
                    file_size=1000,
                    width=800,
                    height=600,
                ),
            ],
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.text == "/file put incoming/photo.jpg"
    assert msg.document is not None
    assert msg.document.file_id == "large"
    assert msg.document.file_name is None
    assert msg.document.file_size == 1000


def test_parse_incoming_update_media_group_id() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            chat=Chat(id=123, type="private"),
            media_group_id="group-1",
            photo=[
                PhotoSize(
                    file_id="large",
                    file_size=1000,
                    width=800,
                    height=600,
                )
            ],
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.media_group_id == "group-1"


def test_parse_incoming_update_video_message() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            caption="/file put incoming/video.mp4",
            chat=Chat(id=123, type="private"),
            video=Video(
                file_id="video-id",
                file_name="video.mp4",
                mime_type="video/mp4",
                file_size=4242,
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.text == "/file put incoming/video.mp4"
    assert msg.document is not None
    assert msg.document.file_id == "video-id"
    assert msg.document.file_name == "video.mp4"
    assert msg.document.mime_type == "video/mp4"
    assert msg.document.file_size == 4242


def test_parse_incoming_update_sticker_message() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            caption="/file put incoming/sticker.webp",
            chat=Chat(id=123, type="private"),
            sticker=Sticker(
                file_id="sticker-id",
                file_size=2468,
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert msg is not None
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.text == "/file put incoming/sticker.webp"
    assert msg.document is not None
    assert msg.document.file_id == "sticker-id"
    assert msg.document.file_name is None
    assert msg.document.mime_type is None
    assert msg.document.file_size == 2468


def test_parse_incoming_update_callback_query() -> None:
    update = Update(
        update_id=1,
        callback_query=CallbackQuery(
            id="cbq-1",
            data="yee88:cancel",
            from_=User(id=321),
            message=CallbackQueryMessage(
                message_id=55,
                chat=Chat(id=123, type="private"),
            ),
        ),
    )

    msg = parse_incoming_update(update, chat_id=123)
    assert isinstance(msg, TelegramCallbackQuery)
    assert msg.transport == "telegram"
    assert msg.chat_id == 123
    assert msg.message_id == 55
    assert msg.callback_query_id == "cbq-1"
    assert msg.data == "yee88:cancel"
    assert msg.sender_id == 321


def test_parse_incoming_update_topic_fields() -> None:
    update = Update(
        update_id=1,
        message=Message(
            message_id=10,
            text="hello",
            message_thread_id=77,
            is_topic_message=True,
            chat=Chat(id=-100, type="supergroup", is_forum=True),
        ),
    )

    msg = parse_incoming_update(update, chat_id=-100)
    assert isinstance(msg, TelegramIncomingMessage)
    assert msg.thread_id == 77
    assert msg.is_topic_message is True
    assert msg.chat_type == "supergroup"
    assert msg.is_forum is True



from __future__ import annotations

from typing import TYPE_CHECKING

from ..context import RunContext
from ..transport_runtime import TransportRuntime
from .topic_state import TopicThreadSnapshot
from .topics import _topics_scope_label

if TYPE_CHECKING:
    from .bridge import TelegramBridgeConfig

__all__ = [
    "_format_context",
    "_format_ctx_status",
    "_merge_topic_context",
    "_parse_project_branch_args",
    "_usage_ctx_set",
    "_usage_topic",
]


def _format_context(runtime: TransportRuntime, context: RunContext | None) -> str:
    if context is None or context.project is None:
        return "none"
    project = runtime.project_alias_for_key(context.project)
    if context.branch:
        return f"{project} @{context.branch}"
    return project


def _usage_ctx_set(*, chat_project: str | None) -> str:
    if chat_project is not None:
        return "usage: `/ctx set [@branch]`"
    return "usage: `/ctx set <project> [@branch]`"


def _usage_topic(*, chat_project: str | None) -> str:
    if chat_project is not None:
        return "usage: `/topic @branch`"
    return "usage: `/topic <project> @branch`"


def _parse_project_branch_args(
    args_text: str,
    *,
    runtime: TransportRuntime,
    require_branch: bool,
    chat_project: str | None,
) -> tuple[RunContext | None, str | None]:
    from .files import split_command_args

    tokens = split_command_args(args_text)
    if not tokens:
        return (
            None,
            _usage_topic(chat_project=chat_project)
            if require_branch
            else _usage_ctx_set(chat_project=chat_project),
        )
    if len(tokens) > 2:
        return None, "too many arguments"
    project_token: str | None = None
    branch: str | None = None
    first = tokens[0]
    if first.startswith("@"):
        branch = first[1:] or None
    else:
        project_token = first
        if len(tokens) == 2:
            second = tokens[1]
            if not second.startswith("@"):
                return None, "branch must be prefixed with @"
            branch = second[1:] or None

    project_key: str | None = None
    if chat_project is not None:
        if project_token is None:
            project_key = chat_project
        else:
            normalized = runtime.normalize_project_key(project_token)
            if normalized is None:
                return None, f"unknown project {project_token!r}"
            if normalized != chat_project:
                expected = runtime.project_alias_for_key(chat_project)
                return None, (f"project mismatch for this chat; expected {expected!r}.")
            project_key = normalized
    else:
        if project_token is None:
            return None, "project is required"
        project_key = runtime.normalize_project_key(project_token)
        if project_key is None:
            return None, f"unknown project {project_token!r}"

    if require_branch and not branch:
        return None, "branch is required"

    return RunContext(project=project_key, branch=branch), None


def _format_ctx_status(
    *,
    cfg: TelegramBridgeConfig,
    runtime: TransportRuntime,
    bound: RunContext | None,
    resolved: RunContext | None,
    context_source: str,
    snapshot: TopicThreadSnapshot | None,
    chat_project: str | None,
    engine: str | None = None,
    model: str | None = None,
) -> str:
    lines = [
        f"topics: enabled (scope={_topics_scope_label(cfg)})",
        f"bound ctx: {_format_context(runtime, bound)}",
        f"resolved ctx: {_format_context(runtime, resolved)} (source: {context_source})",
    ]
    if chat_project is None and bound is None:
        topic_usage = (
            _usage_topic(chat_project=chat_project).removeprefix("usage: ").strip()
        )
        ctx_usage = (
            _usage_ctx_set(chat_project=chat_project).removeprefix("usage: ").strip()
        )
        lines.append(f"note: unbound topic — bind with {topic_usage} or {ctx_usage}")
    sessions = None
    if snapshot is not None and snapshot.sessions:
        sessions = ", ".join(sorted(snapshot.sessions))
    lines.append(f"sessions: {sessions or 'none'}")
    if engine is not None:
        lines.append(f"engine: {engine}")
    if model is not None:
        lines.append(f"model: {model}")
    return "\n".join(lines)


def _merge_topic_context(
    *, chat_project: str | None, bound: RunContext | None
) -> RunContext | None:
    if chat_project is None:
        return bound
    if bound is None:
        return RunContext(project=chat_project, branch=None)
    if bound.project is None:
        return RunContext(project=chat_project, branch=bound.branch)
    return bound



from __future__ import annotations

from typing import TYPE_CHECKING

from ...commands import get_command
from ...config import ConfigError
from ...ids import RESERVED_COMMAND_IDS, is_valid_id
from ...logging import get_logger
from ...plugins import COMMAND_GROUP, list_entrypoints
from ...transport_runtime import TransportRuntime

if TYPE_CHECKING:
    from ..bridge import TelegramBridgeConfig

logger = get_logger(__name__)

_MAX_BOT_COMMANDS = 100


def build_bot_commands(
    runtime: TransportRuntime,
    *,
    include_file: bool = True,
    include_topics: bool = False,
) -> list[dict[str, str]]:
    commands: list[dict[str, str]] = []
    seen: set[str] = set()
    for engine_id in runtime.available_engine_ids():
        cmd = engine_id.lower()
        if cmd in seen:
            continue
        commands.append({"command": cmd, "description": f"use engine: {cmd}"})
        seen.add(cmd)
    for alias in runtime.project_aliases():
        cmd = alias.lower()
        if cmd in seen:
            continue
        if not is_valid_id(cmd):
            logger.debug(
                "startup.command_menu.skip_project",
                alias=alias,
            )
            continue
        commands.append({"command": cmd, "description": f"work on: {cmd}"})
        seen.add(cmd)
    allowlist = runtime.allowlist
    for ep in list_entrypoints(
        COMMAND_GROUP,
        allowlist=allowlist,
        reserved_ids=RESERVED_COMMAND_IDS,
    ):
        try:
            backend = get_command(ep.name, allowlist=allowlist)
        except ConfigError as exc:
            logger.info(
                "startup.command_menu.skip_command",
                command=ep.name,
                error=str(exc),
            )
            continue
        cmd = backend.id.lower()
        if cmd in seen:
            continue
        if not is_valid_id(cmd):
            logger.debug(
                "startup.command_menu.skip_command_id",
                command=cmd,
            )
            continue
        description = backend.description or f"command: {cmd}"
        commands.append({"command": cmd, "description": description})
        seen.add(cmd)
    for cmd, description in [
        ("new", "start a new thread"),
        ("ctx", "show or update context"),
        ("agent", "set default engine"),
        ("model", "set model override"),
        ("reasoning", "set reasoning override"),
        ("trigger", "set trigger mode"),
    ]:
        if cmd in seen:
            continue
        commands.append({"command": cmd, "description": description})
        seen.add(cmd)
    if include_topics:
        for cmd, description in [
            ("topic", "create or bind a topic"),
            ("fork", "fork current topic to new topic"),
        ]:
            if cmd in seen:
                continue
            commands.append({"command": cmd, "description": description})
            seen.add(cmd)
    if include_file and "file" not in seen:
        commands.append({"command": "file", "description": "upload or fetch files"})
        seen.add("file")
    if "cancel" not in seen:
        commands.append({"command": "cancel", "description": "cancel run"})
    if len(commands) > _MAX_BOT_COMMANDS:
        logger.warning(
            "startup.command_menu.too_many",
            count=len(commands),
            limit=_MAX_BOT_COMMANDS,
        )
        commands = commands[:_MAX_BOT_COMMANDS]
        if not any(cmd["command"] == "cancel" for cmd in commands):
            commands[-1] = {"command": "cancel", "description": "cancel run"}
    return commands


def _reserved_commands(runtime: TransportRuntime) -> set[str]:
    return {
        *{engine.lower() for engine in runtime.engine_ids},
        *{alias.lower() for alias in runtime.project_aliases()},
        *RESERVED_COMMAND_IDS,
    }


async def _set_command_menu(cfg: TelegramBridgeConfig) -> None:
    commands = build_bot_commands(
        cfg.runtime,
        include_file=cfg.files.enabled,
        include_topics=cfg.topics.enabled,
    )
    if not commands:
        return
    try:
        ok = await cfg.bot.set_my_commands(commands)
    except Exception as exc:  # noqa: BLE001
        logger.info(
            "startup.command_menu.failed",
            error=str(exc),
            error_type=exc.__class__.__name__,
        )
        return
    if not ok:
        logger.info("startup.command_menu.rejected")
        return
    logger.info(
        "startup.command_menu.updated",
        commands=[cmd["command"] for cmd in commands],
    )



from __future__ import annotations

import os
from contextvars import ContextVar, Token
from pathlib import Path


_run_base_dir: ContextVar[Path | None] = ContextVar("yee88_run_base_dir", default=None)


def get_run_base_dir() -> Path | None:
    return _run_base_dir.get()


def set_run_base_dir(base_dir: Path | None) -> Token[Path | None]:
    return _run_base_dir.set(base_dir)


def reset_run_base_dir(token: Token[Path | None]) -> None:
    _run_base_dir.reset(token)


def relativize_path(value: str, *, base_dir: Path | None = None) -> str:
    if not value:
        return value
    base = get_run_base_dir() if base_dir is None else base_dir
    if base is None:
        base = Path.cwd()
    base_str = str(base)
    if not base_str:
        return value
    if value == base_str:
        return "."
    for sep in (os.sep, "/"):
        prefix = base_str if base_str.endswith(sep) else f"{base_str}{sep}"
        if value.startswith(prefix):
            suffix = value[len(prefix) :]
            return suffix or "."
    return value


def relativize_command(value: str, *, base_dir: Path | None = None) -> str:
    base = get_run_base_dir() if base_dir is None else base_dir
    if base is None:
        base = Path.cwd()
    base_with_sep = f"{base}{os.sep}"
    return value.replace(base_with_sep, "")



from __future__ import annotations

from pathlib import Path

import pytest

from yee88.schemas import opencode as opencode_schema


def _fixture_path(name: str) -> Path:
    return Path(__file__).parent / "fixtures" / name


def _decode_fixture(name: str) -> list[str]:
    path = _fixture_path(name)
    errors: list[str] = []

    for lineno, line in enumerate(path.read_bytes().splitlines(), 1):
        if not line.strip():
            continue
        try:
            opencode_schema.decode_event(line)
        except Exception as exc:  # noqa: BLE001
            errors.append(f"line {lineno}: {exc.__class__.__name__}: {exc}")

    return errors


@pytest.mark.parametrize(
    "fixture",
    [
        "opencode_stream_success.jsonl",
        "opencode_stream_success_no_reason.jsonl",
        "opencode_stream_error.jsonl",
        "opencode_run_json.jsonl",
    ],
)
def test_opencode_schema_parses_fixture(fixture: str) -> None:
    errors = _decode_fixture(fixture)
    assert not errors, f"{fixture} had {len(errors)} errors: " + "; ".join(errors[:5])



from __future__ import annotations

import sys
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any, cast

import anyio
import typer

from ..config import ConfigError, load_or_init_config, write_config
from ..config_migrations import migrate_config
from ..logging import setup_logging
from ..settings import TakopiSettings
from ..telegram import onboarding
from .init import _ensure_projects_table
from .run import _load_settings_optional


def chat_id(
    token: str | None = typer.Option(
        None,
        "--token",
        help="Telegram bot token (defaults to config if available).",
    ),
    project: str | None = typer.Option(
        None,
        "--project",
        help="Project alias to print a chat_id snippet for.",
    ),
) -> None:
    """Capture a Telegram chat id and exit."""
    setup_logging_fn = cast(
        Callable[..., None],
        _resolve_cli_attr("setup_logging") or setup_logging,
    )
    load_settings_optional_fn = cast(
        Callable[[], tuple[TakopiSettings | None, Path | None]],
        _resolve_cli_attr("_load_settings_optional") or _load_settings_optional,
    )
    onboarding_mod = cast(
        Any,
        _resolve_cli_attr("onboarding") or onboarding,
    )
    load_or_init_config_fn = cast(
        Callable[[], tuple[dict, Path]],
        _resolve_cli_attr("load_or_init_config") or load_or_init_config,
    )
    ensure_projects_table_fn = cast(
        Callable[[dict, Path], dict],
        _resolve_cli_attr("_ensure_projects_table") or _ensure_projects_table,
    )
    migrate_config_fn = cast(
        Callable[..., object],
        _resolve_cli_attr("migrate_config") or migrate_config,
    )
    write_config_fn = cast(
        Callable[[dict, Path], None],
        _resolve_cli_attr("write_config") or write_config,
    )

    setup_logging_fn(debug=False, cache_logger_on_first_use=False)
    if token is None:
        settings, _ = load_settings_optional_fn()
        if settings is not None:
            tg = settings.transports.telegram
            token = tg.bot_token or None
    chat = anyio.run(partial(onboarding_mod.capture_chat_id, token=token))
    if chat is None:
        raise typer.Exit(code=1)
    if project:
        project = project.strip()
        if not project:
            raise ConfigError("Invalid `--project`; expected a non-empty string.")

        config, config_path = load_or_init_config_fn()
        if config_path.exists():
            applied = migrate_config_fn(config, config_path=config_path)
            if applied:
                write_config_fn(config, config_path)

        projects = ensure_projects_table_fn(config, config_path)
        entry = projects.get(project)
        if entry is None:
            lowered = project.lower()
            for key, value in projects.items():
                if isinstance(key, str) and key.lower() == lowered:
                    entry = value
                    project = key
                    break
        if entry is None:
            raise ConfigError(
                f"Unknown project {project!r}; run `yee88 init {project}` first."
            )
        if not isinstance(entry, dict):
            raise ConfigError(
                f"Invalid `projects.{project}` in {config_path}; expected a table."
            )
        entry["chat_id"] = chat.chat_id
        write_config_fn(config, config_path)
        typer.echo(f"updated projects.{project}.chat_id = {chat.chat_id}")
        return

    typer.echo(f"chat_id = {chat.chat_id}")


def onboarding_paths() -> None:
    """Print all possible onboarding paths."""
    setup_logging_fn = cast(
        Callable[..., None],
        _resolve_cli_attr("setup_logging") or setup_logging,
    )
    onboarding_mod = cast(
        Any,
        _resolve_cli_attr("onboarding") or onboarding,
    )
    setup_logging_fn(debug=False, cache_logger_on_first_use=False)
    onboarding_mod.debug_onboarding_paths()


def _resolve_cli_attr(name: str) -> object | None:
    cli_module = sys.modules.get("yee88.cli")
    if cli_module is None:
        return None
    return getattr(cli_module, name, None)



import json
from pathlib import Path
from typing import cast

import anyio
import pytest

import yee88.runners.claude as claude_runner
from yee88.model import ActionEvent, CompletedEvent, ResumeToken, StartedEvent
from yee88.runners.claude import (
    ClaudeRunner,
    ClaudeStreamState,
    ENGINE,
    translate_claude_event,
)
from yee88.schemas import claude as claude_schema


def _load_fixture(
    name: str, *, session_id: str | None = None
) -> list[claude_schema.StreamJsonMessage]:
    path = Path(__file__).parent / "fixtures" / name
    events = [
        claude_schema.decode_stream_json_line(line)
        for line in path.read_bytes().splitlines()
        if line.strip()
    ]
    if session_id is None:
        return events
    return [
        event for event in events if getattr(event, "session_id", None) == session_id
    ]


def _decode_event(payload: dict) -> claude_schema.StreamJsonMessage:
    data_payload = dict(payload)
    data_payload.setdefault("uuid", "uuid")
    data_payload.setdefault("session_id", "session")
    match data_payload.get("type"):
        case "assistant":
            message = dict(data_payload.get("message", {}))
            message.setdefault("role", "assistant")
            message.setdefault("content", [])
            message.setdefault("model", "claude")
            data_payload["message"] = message
        case "user":
            message = dict(data_payload.get("message", {}))
            message.setdefault("role", "user")
            message.setdefault("content", [])
            data_payload["message"] = message
    data = json.dumps(data_payload).encode("utf-8")
    return claude_schema.decode_stream_json_line(data)


def test_claude_resume_format_and_extract() -> None:
    runner = ClaudeRunner(claude_cmd="claude")
    token = ResumeToken(engine=ENGINE, value="sid")

    assert runner.format_resume(token) == "`claude --resume sid`"
    assert runner.extract_resume("`claude --resume sid`") == token
    assert runner.extract_resume("claude -r other") == ResumeToken(
        engine=ENGINE, value="other"
    )
    assert runner.extract_resume("`codex resume sid`") is None


def test_build_runner_uses_shutil_which(monkeypatch) -> None:
    expected = r"C:\Tools\claude.cmd"
    called: dict[str, str] = {}

    def fake_which(name: str) -> str | None:
        called["name"] = name
        return expected

    monkeypatch.setattr(claude_runner.shutil, "which", fake_which)
    runner = cast(ClaudeRunner, claude_runner.build_runner({}, Path("yee88.toml")))

    assert called["name"] == "claude"
    assert runner.claude_cmd == expected


def test_translate_success_fixture() -> None:
    state = ClaudeStreamState()
    events: list = []
    for event in _load_fixture(
        "claude_stream_json_session.jsonl",
        session_id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
    ):
        events.extend(
            translate_claude_event(
                event,
                title="claude",
                state=state,
                factory=state.factory,
            )
        )

    assert isinstance(events[0], StartedEvent)
    started = next(evt for evt in events if isinstance(evt, StartedEvent))

    action_events = [evt for evt in events if isinstance(evt, ActionEvent)]
    assert len(action_events) == 4

    started_actions = {
        (evt.action.id, evt.phase): evt
        for evt in action_events
        if evt.phase == "started"
    }
    assert (
        started_actions[("toolu_01BASH_LS_EXAMPLE", "started")].action.kind == "command"
    )
    write_action = started_actions[("toolu_02", "started")].action
    assert write_action.kind == "file_change"
    assert write_action.detail["changes"][0]["path"] == "notes.md"

    completed_actions = {
        (evt.action.id, evt.phase): evt
        for evt in action_events
        if evt.phase == "completed"
    }
    assert completed_actions[("toolu_01BASH_LS_EXAMPLE", "completed")].ok is True
    assert completed_actions[("toolu_02", "completed")].ok is True

    completed = next(evt for evt in events if isinstance(evt, CompletedEvent))
    assert events[-1] == completed
    assert completed.ok is True
    assert completed.resume == started.resume
    assert completed.answer == "I see README.md, pyproject.toml, and src/."


def test_translate_error_fixture_permission_denials() -> None:
    state = ClaudeStreamState()
    events: list = []
    for event in _load_fixture(
        "claude_stream_json_session.jsonl",
        session_id="bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb",
    ):
        events.extend(
            translate_claude_event(
                event,
                title="claude",
                state=state,
                factory=state.factory,
            )
        )

    started = next(evt for evt in events if isinstance(evt, StartedEvent))
    completed = next(evt for evt in events if isinstance(evt, CompletedEvent))
    assert completed.ok is False
    assert completed.error is not None
    assert "claude run failed" in completed.error
    assert completed.resume == started.resume


def test_tool_results_pop_pending_actions() -> None:
    state = ClaudeStreamState()

    tool_use_event = {
        "type": "assistant",
        "message": {
            "id": "msg_1",
            "content": [
                {
                    "type": "tool_use",
                    "id": "toolu_1",
                    "name": "Bash",
                    "input": {"command": "echo hi"},
                }
            ],
        },
    }
    tool_result_event = {
        "type": "user",
        "message": {
            "id": "msg_2",
            "content": [
                {
                    "type": "tool_result",
                    "tool_use_id": "toolu_1",
                    "content": "ok",
                    "is_error": False,
                }
            ],
        },
    }

    translate_claude_event(
        _decode_event(tool_use_event),
        title="claude",
        state=state,
        factory=state.factory,
    )
    assert "toolu_1" in state.pending_actions

    translate_claude_event(
        _decode_event(tool_result_event),
        title="claude",
        state=state,
        factory=state.factory,
    )
    assert not state.pending_actions


def test_translate_thinking_block() -> None:
    state = ClaudeStreamState()
    event = {
        "type": "assistant",
        "message": {
            "id": "msg_1",
            "content": [
                {
                    "type": "thinking",
                    "thinking": "Consider the options.",
                    "signature": "sig",
                }
            ],
        },
    }

    events = translate_claude_event(
        _decode_event(event),
        title="claude",
        state=state,
        factory=state.factory,
    )

    assert len(events) == 1
    assert isinstance(events[0], ActionEvent)
    assert events[0].phase == "completed"
    assert events[0].action.kind == "note"
    assert events[0].action.title == "Consider the options."
    assert events[0].ok is True


@pytest.mark.anyio
async def test_run_serializes_same_session() -> None:
    runner = ClaudeRunner(claude_cmd="claude")
    gate = anyio.Event()
    in_flight = 0
    max_in_flight = 0

    async def run_stub(*_args, **_kwargs):
        nonlocal in_flight, max_in_flight
        in_flight += 1
        max_in_flight = max(max_in_flight, in_flight)
        try:
            await gate.wait()
            yield CompletedEvent(
                engine=ENGINE,
                resume=ResumeToken(engine=ENGINE, value="sid"),
                ok=True,
                answer="ok",
            )
        finally:
            in_flight -= 1

    runner.run_impl = run_stub  # type: ignore[assignment]

    async def drain(prompt: str, resume: ResumeToken | None) -> None:
        async for _event in runner.run(prompt, resume):
            pass

    token = ResumeToken(engine=ENGINE, value="sid")
    async with anyio.create_task_group() as tg:
        tg.start_soon(drain, "a", token)
        tg.start_soon(drain, "b", token)
        await anyio.sleep(0)
        gate.set()
    assert max_in_flight == 1


@pytest.mark.anyio
async def test_run_serializes_new_session_after_session_is_known(
    tmp_path, monkeypatch
) -> None:
    gate_path = tmp_path / "gate"
    resume_marker = tmp_path / "resume_started"
    session_id = "session_01"

    claude_path = tmp_path / "claude"
    claude_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import os\n"
        "import sys\n"
        "import time\n"
        "\n"
        "gate = os.environ['CLAUDE_TEST_GATE']\n"
        "resume_marker = os.environ['CLAUDE_TEST_RESUME_MARKER']\n"
        "session_id = os.environ['CLAUDE_TEST_SESSION_ID']\n"
        "\n"
        "init = {\n"
        "    'type': 'system',\n"
        "    'subtype': 'init',\n"
        "    'uuid': 'uuid',\n"
        "    'session_id': session_id,\n"
        "    'apiKeySource': 'env',\n"
        "    'cwd': '.',\n"
        "    'tools': [],\n"
        "    'mcp_servers': [],\n"
        "    'model': 'claude',\n"
        "    'permissionMode': 'default',\n"
        "    'slash_commands': [],\n"
        "    'output_style': 'default',\n"
        "}\n"
        "\n"
        "args = sys.argv[1:]\n"
        "if '--resume' in args or '-r' in args:\n"
        "    print(json.dumps(init), flush=True)\n"
        "    with open(resume_marker, 'w', encoding='utf-8') as f:\n"
        "        f.write('started')\n"
        "        f.flush()\n"
        "    sys.exit(0)\n"
        "\n"
        "print(json.dumps(init), flush=True)\n"
        "while not os.path.exists(gate):\n"
        "    time.sleep(0.001)\n"
        "sys.exit(0)\n",
        encoding="utf-8",
    )
    claude_path.chmod(0o755)

    monkeypatch.setenv("CLAUDE_TEST_GATE", str(gate_path))
    monkeypatch.setenv("CLAUDE_TEST_RESUME_MARKER", str(resume_marker))
    monkeypatch.setenv("CLAUDE_TEST_SESSION_ID", session_id)

    runner = ClaudeRunner(claude_cmd=str(claude_path))

    session_started = anyio.Event()
    resume_value: str | None = None
    new_done = anyio.Event()

    async def run_new() -> None:
        nonlocal resume_value
        async for event in runner.run("hello", None):
            if isinstance(event, StartedEvent):
                resume_value = event.resume.value
                session_started.set()
        new_done.set()

    async def run_resume() -> None:
        assert resume_value is not None
        async for _event in runner.run(
            "resume", ResumeToken(engine=ENGINE, value=resume_value)
        ):
            pass

    async with anyio.create_task_group() as tg:
        tg.start_soon(run_new)
        await session_started.wait()

        tg.start_soon(run_resume)
        await anyio.sleep(0.01)

        assert not resume_marker.exists()

        gate_path.write_text("go", encoding="utf-8")
        await new_done.wait()

        with anyio.fail_after(2):
            while not resume_marker.exists():
                await anyio.sleep(0.001)


@pytest.mark.anyio
async def test_run_strips_anthropic_api_key_by_default(tmp_path, monkeypatch) -> None:
    claude_path = tmp_path / "claude"
    claude_path.write_text(
        "#!/usr/bin/env python3\n"
        "import json\n"
        "import os\n"
        "\n"
        "session_id = 'session_01'\n"
        "status = 'set' if os.environ.get('ANTHROPIC_API_KEY') else 'unset'\n"
        "init = {\n"
        "    'type': 'system',\n"
        "    'subtype': 'init',\n"
        "    'uuid': 'uuid',\n"
        "    'session_id': session_id,\n"
        "    'apiKeySource': 'env',\n"
        "    'cwd': '.',\n"
        "    'tools': [],\n"
        "    'mcp_servers': [],\n"
        "    'model': 'claude',\n"
        "    'permissionMode': 'default',\n"
        "    'slash_commands': [],\n"
        "    'output_style': 'default',\n"
        "}\n"
        "print(json.dumps(init), flush=True)\n"
        "result = {\n"
        "    'type': 'result',\n"
        "    'subtype': 'success',\n"
        "    'uuid': 'uuid',\n"
        "    'session_id': session_id,\n"
        "    'duration_ms': 0,\n"
        "    'duration_api_ms': 0,\n"
        "    'is_error': False,\n"
        "    'num_turns': 1,\n"
        "    'result': f'api={status}',\n"
        "    'total_cost_usd': 0.0,\n"
        "    'usage': {'input_tokens': 0, 'output_tokens': 0},\n"
        "    'modelUsage': {},\n"
        "    'permission_denials': [],\n"
        "}\n"
        "print(json.dumps(result), flush=True)\n"
        "raise SystemExit(0)\n",
        encoding="utf-8",
    )
    claude_path.chmod(0o755)

    monkeypatch.setenv("ANTHROPIC_API_KEY", "secret")

    runner = ClaudeRunner(claude_cmd=str(claude_path))
    answer: str | None = None
    async for event in runner.run("hello", None):
        if isinstance(event, CompletedEvent):
            answer = event.answer
    assert answer == "api=unset"

    runner_api = ClaudeRunner(claude_cmd=str(claude_path), use_api_billing=True)
    answer = None
    async for event in runner_api.run("hello", None):
        if isinstance(event, CompletedEvent):
            answer = event.answer
    assert answer == "api=set"



from __future__ import annotations

from pathlib import Path

from .config import ProjectConfig, ProjectsConfig
from .context import RunContext
from .utils.git import (
    git_is_worktree,
    git_ok,
    git_run,
    git_stdout,
    resolve_default_base,
)


class WorktreeError(RuntimeError):
    pass


def resolve_run_cwd(
    context: RunContext | None,
    *,
    projects: ProjectsConfig,
) -> Path | None:
    if context is None or context.project is None:
        return None
    project = projects.projects.get(context.project)
    if project is None:
        raise WorktreeError(f"unknown project {context.project!r}")
    if context.branch is None:
        return project.path
    branch = _sanitize_branch(context.branch)
    if _matches_project_branch(project.path, branch):
        return project.path
    if project.worktrees_dir is None:
        return project.path
    return ensure_worktree(project, branch)


def ensure_worktree(project: ProjectConfig, branch: str) -> Path:
    root = project.path
    if not root.exists():
        raise WorktreeError(f"project path not found: {root}")

    branch = _sanitize_branch(branch)
    worktrees_root = project.worktrees_root
    if worktrees_root is None:
        raise WorktreeError(
            f"project {project.alias!r} has no worktrees_dir configured"
        )
    assert isinstance(worktrees_root, Path)
    worktree_path = worktrees_root / branch
    _ensure_within_root(worktrees_root, worktree_path)

    if worktree_path.exists():
        if not git_is_worktree(worktree_path):
            raise WorktreeError(f"{worktree_path} exists but is not a git worktree")
        return worktree_path

    worktrees_root.mkdir(parents=True, exist_ok=True)

    if git_ok(
        ["show-ref", "--verify", "--quiet", f"refs/heads/{branch}"],
        cwd=root,
    ):
        _git_worktree_add(root, worktree_path, branch)
        return worktree_path

    if git_ok(
        ["show-ref", "--verify", "--quiet", f"refs/remotes/origin/{branch}"],
        cwd=root,
    ):
        _git_worktree_add(
            root,
            worktree_path,
            branch,
            base_ref=f"origin/{branch}",
            create_branch=True,
        )
        return worktree_path

    base = project.worktree_base or resolve_default_base(root)
    if not base:
        raise WorktreeError("cannot determine base branch for new worktree")

    _git_worktree_add(
        root,
        worktree_path,
        branch,
        base_ref=base,
        create_branch=True,
    )
    return worktree_path


def _git_worktree_add(
    root: Path,
    worktree_path: Path,
    branch: str,
    *,
    base_ref: str | None = None,
    create_branch: bool = False,
) -> None:
    if create_branch:
        if not base_ref:
            raise WorktreeError("missing base ref for worktree creation")
        args = ["worktree", "add", "-b", branch, str(worktree_path), base_ref]
    else:
        args = ["worktree", "add", str(worktree_path), branch]

    result = git_run(args, cwd=root)
    if result is None:
        raise WorktreeError("git not available on PATH")
    if result.returncode != 0:
        message = result.stderr.strip() or result.stdout.strip()
        raise WorktreeError(message or "git worktree add failed")


def _sanitize_branch(branch: str) -> str:
    cleaned = branch.strip()
    if not cleaned:
        raise WorktreeError("branch name cannot be empty")
    if cleaned.startswith("/"):
        raise WorktreeError("branch name cannot start with '/'")
    for part in Path(cleaned).parts:
        if part == "..":
            raise WorktreeError("branch name cannot contain '..'")
    return cleaned


def _matches_project_branch(root: Path, branch: str) -> bool:
    current = git_stdout(["branch", "--show-current"], cwd=root)
    if not current:
        return False
    return current == branch


def _ensure_within_root(root: Path, path: Path) -> None:
    root_resolved = root.resolve(strict=False)
    path_resolved = path.resolve(strict=False)
    if not path_resolved.is_relative_to(root_resolved):
        raise WorktreeError("branch path escapes the worktrees directory")



"""Msgspec models and decoder for Claude Code stream-json output."""

from __future__ import annotations

from typing import Any, Literal

import msgspec


class StreamTextBlock(
    msgspec.Struct, tag="text", tag_field="type", forbid_unknown_fields=False
):
    text: str


class StreamThinkingBlock(
    msgspec.Struct, tag="thinking", tag_field="type", forbid_unknown_fields=False
):
    thinking: str
    signature: str


class StreamToolUseBlock(
    msgspec.Struct, tag="tool_use", tag_field="type", forbid_unknown_fields=False
):
    id: str
    name: str
    input: dict[str, Any]


class StreamToolResultBlock(
    msgspec.Struct, tag="tool_result", tag_field="type", forbid_unknown_fields=False
):
    tool_use_id: str
    content: str | list[dict[str, Any]] | None = None
    is_error: bool | None = None


type StreamContentBlock = (
    StreamTextBlock | StreamThinkingBlock | StreamToolUseBlock | StreamToolResultBlock
)


class StreamUserMessageBody(msgspec.Struct, forbid_unknown_fields=False):
    role: Literal["user"]
    content: str | list[StreamContentBlock]


class StreamAssistantMessageBody(msgspec.Struct, forbid_unknown_fields=False):
    role: Literal["assistant"]
    content: list[StreamContentBlock]
    model: str
    error: str | None = None


class StreamUserMessage(
    msgspec.Struct, tag="user", tag_field="type", forbid_unknown_fields=False
):
    message: StreamUserMessageBody
    uuid: str | None = None
    parent_tool_use_id: str | None = None
    session_id: str | None = None


class StreamAssistantMessage(
    msgspec.Struct, tag="assistant", tag_field="type", forbid_unknown_fields=False
):
    message: StreamAssistantMessageBody
    parent_tool_use_id: str | None = None
    uuid: str | None = None
    session_id: str | None = None


class StreamSystemMessage(
    msgspec.Struct, tag="system", tag_field="type", forbid_unknown_fields=False
):
    subtype: str
    session_id: str | None = None
    uuid: str | None = None
    cwd: str | None = None
    tools: list[str] | None = None
    mcp_servers: list[Any] | None = None
    model: str | None = None
    permissionMode: str | None = None
    output_style: str | None = None
    apiKeySource: str | None = None


class StreamResultMessage(
    msgspec.Struct, tag="result", tag_field="type", forbid_unknown_fields=False
):
    subtype: str
    duration_ms: int
    duration_api_ms: int
    is_error: bool
    num_turns: int
    session_id: str
    total_cost_usd: float | None = None
    usage: dict[str, Any] | None = None
    result: str | None = None
    structured_output: Any = None


class StreamEventMessage(
    msgspec.Struct, tag="stream_event", tag_field="type", forbid_unknown_fields=False
):
    uuid: str
    session_id: str
    event: dict[str, Any]
    parent_tool_use_id: str | None = None


class ControlInterruptRequest(
    msgspec.Struct, tag="interrupt", tag_field="subtype", forbid_unknown_fields=False
):
    pass


class ControlCanUseToolRequest(
    msgspec.Struct, tag="can_use_tool", tag_field="subtype", forbid_unknown_fields=False
):
    tool_name: str
    input: dict[str, Any]
    permission_suggestions: list[Any] | None = None
    blocked_path: str | None = None


class ControlInitializeRequest(
    msgspec.Struct, tag="initialize", tag_field="subtype", forbid_unknown_fields=False
):
    hooks: dict[str, Any] | None = None


class ControlSetPermissionModeRequest(
    msgspec.Struct,
    tag="set_permission_mode",
    tag_field="subtype",
    forbid_unknown_fields=False,
):
    mode: str


class ControlHookCallbackRequest(
    msgspec.Struct,
    tag="hook_callback",
    tag_field="subtype",
    forbid_unknown_fields=False,
):
    callback_id: str
    input: Any
    tool_use_id: str | None = None


class ControlMcpMessageRequest(
    msgspec.Struct, tag="mcp_message", tag_field="subtype", forbid_unknown_fields=False
):
    server_name: str
    message: Any


class ControlRewindFilesRequest(
    msgspec.Struct, tag="rewind_files", tag_field="subtype", forbid_unknown_fields=False
):
    user_message_id: str


type ControlRequest = (
    ControlInterruptRequest
    | ControlCanUseToolRequest
    | ControlInitializeRequest
    | ControlSetPermissionModeRequest
    | ControlHookCallbackRequest
    | ControlMcpMessageRequest
    | ControlRewindFilesRequest
)


class StreamControlRequest(
    msgspec.Struct, tag="control_request", tag_field="type", forbid_unknown_fields=False
):
    request_id: str
    request: ControlRequest


class ControlSuccessResponse(
    msgspec.Struct, tag="success", tag_field="subtype", forbid_unknown_fields=False
):
    request_id: str
    response: dict[str, Any] | None = None


class ControlErrorResponse(
    msgspec.Struct, tag="error", tag_field="subtype", forbid_unknown_fields=False
):
    request_id: str
    error: str


type ControlResponse = ControlSuccessResponse | ControlErrorResponse


class StreamControlResponse(
    msgspec.Struct,
    tag="control_response",
    tag_field="type",
    forbid_unknown_fields=False,
):
    response: ControlResponse


class StreamControlCancelRequest(
    msgspec.Struct,
    tag="control_cancel_request",
    tag_field="type",
    forbid_unknown_fields=False,
):
    request_id: str | None = None


type StreamJsonMessage = (
    StreamUserMessage
    | StreamAssistantMessage
    | StreamSystemMessage
    | StreamResultMessage
    | StreamEventMessage
    | StreamControlRequest
    | StreamControlResponse
    | StreamControlCancelRequest
)


STREAM_JSON_SCHEMA = msgspec.json.schema(StreamJsonMessage)

_DECODER = msgspec.json.Decoder(StreamJsonMessage)


def decode_stream_json_line(line: str | bytes) -> StreamJsonMessage:
    return _DECODER.decode(line)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from collections.abc import Callable, Iterable


@dataclass(frozen=True, slots=True)
class FakeDist:
    name: str


class FakeEntryPoint:
    def __init__(
        self,
        name: str,
        value: str,
        group: str,
        *,
        loader: Callable[[], Any] | None = None,
        dist_name: str | None = "yee88",
    ) -> None:
        self.name = name
        self.value = value
        self.group = group
        self._loader = loader or (lambda: None)
        self.dist = FakeDist(dist_name) if dist_name else None

    def load(self) -> Any:
        return self._loader()


class FakeEntryPoints(list):
    def select(self, *, group: str) -> list[FakeEntryPoint]:
        return [ep for ep in self if ep.group == group]

    def get(self, group: str, default: Iterable[Any] | None = None) -> list[Any]:
        _ = default
        return [ep for ep in self if ep.group == group]


def install_entrypoints(monkeypatch, entrypoints: Iterable[FakeEntryPoint]) -> None:
    from yee88 import plugins

    def _entry_points() -> FakeEntryPoints:
        return FakeEntryPoints(entrypoints)

    monkeypatch.setattr(plugins, "entry_points", _entry_points)



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import msgspec

from ..backends import EngineBackend, EngineConfig
from ..config import ConfigError
from ..events import EventFactory
from ..logging import get_logger
from ..model import ActionPhase, EngineId, ResumeToken, TakopiEvent
from ..runner import JsonlSubprocessRunner, ResumeTokenMixin, Runner
from .run_options import get_run_options
from ..schemas import codex as codex_schema
from ..utils.paths import relativize_command

logger = get_logger(__name__)

ENGINE: EngineId = "codex"

__all__ = [
    "ENGINE",
    "CodexRunner",
    "find_exec_only_flag",
    "translate_codex_event",
]

_RESUME_RE = re.compile(r"(?im)^\s*`?codex\s+resume\s+(?P<token>[^`\s]+)`?\s*$")
_RECONNECTING_RE = re.compile(
    r"^Reconnecting\.{3}\s*(?P<attempt>\d+)/(?P<max>\d+)\s*$",
    re.IGNORECASE,
)
_EXEC_ONLY_FLAGS = {
    "--skip-git-repo-check",
    "--json",
    "--output-schema",
    "--output-last-message",
    "--color",
    "-o",
}
_EXEC_ONLY_PREFIXES = (
    "--output-schema=",
    "--output-last-message=",
    "--color=",
)


def find_exec_only_flag(extra_args: list[str]) -> str | None:
    for arg in extra_args:
        if arg in _EXEC_ONLY_FLAGS:
            return arg
        for prefix in _EXEC_ONLY_PREFIXES:
            if arg.startswith(prefix):
                return arg
    return None


def _parse_reconnect_message(message: str) -> tuple[int, int] | None:
    match = _RECONNECTING_RE.match(message)
    if not match:
        return None
    try:
        attempt = int(match.group("attempt"))
        max_attempts = int(match.group("max"))
    except (TypeError, ValueError):
        return None
    return (attempt, max_attempts)


def _short_tool_name(server: str | None, tool: str | None) -> str:
    name = ".".join(part for part in (server, tool) if part)
    return name or "tool"


def _summarize_tool_result(result: Any) -> dict[str, Any] | None:
    if isinstance(result, codex_schema.McpToolCallItemResult):
        summary: dict[str, Any] = {}
        content = result.content
        if isinstance(content, list):
            summary["content_blocks"] = len(content)
        elif content is not None:
            summary["content_blocks"] = 1
        summary["has_structured"] = result.structured_content is not None
        return summary or None

    if isinstance(result, dict):
        summary = {}
        content = result.get("content")
        if isinstance(content, list):
            summary["content_blocks"] = len(content)
        elif content is not None:
            summary["content_blocks"] = 1

        structured_key: str | None = None
        if "structured_content" in result:
            structured_key = "structured_content"
        elif "structured" in result:
            structured_key = "structured"

        if structured_key is not None:
            summary["has_structured"] = result.get(structured_key) is not None
        return summary or None

    return None


def _normalize_change_list(changes: list[Any]) -> list[dict[str, str]]:
    normalized: list[dict[str, str]] = []
    for change in changes:
        path: str | None = None
        kind: str | None = None
        if isinstance(change, codex_schema.FileUpdateChange):
            path = change.path
            kind = change.kind
        elif isinstance(change, dict):
            path = change.get("path")
            kind = change.get("kind")
        if not isinstance(path, str) or not path:
            continue
        entry = {"path": path}
        if isinstance(kind, str) and kind:
            entry["kind"] = kind
        normalized.append(entry)
    return normalized


def _format_change_summary(changes: list[Any]) -> str:
    paths: list[str] = []
    for change in changes:
        if isinstance(change, codex_schema.FileUpdateChange):
            if change.path:
                paths.append(change.path)
            continue
        if isinstance(change, dict):
            path = change.get("path")
            if isinstance(path, str) and path:
                paths.append(path)
    if not paths:
        total = len(changes)
        if total <= 0:
            return "files"
        return f"{total} files"
    return ", ".join(str(path) for path in paths)


@dataclass(frozen=True, slots=True)
class _TodoSummary:
    done: int
    total: int
    next_text: str | None


def _summarize_todo_list(items: Any) -> _TodoSummary:
    if not isinstance(items, list):
        return _TodoSummary(done=0, total=0, next_text=None)

    done = 0
    total = 0
    next_text: str | None = None

    for raw_item in items:
        if isinstance(raw_item, codex_schema.TodoItem):
            total += 1
            if raw_item.completed:
                done += 1
                continue
            if next_text is None:
                next_text = raw_item.text
            continue
        if not isinstance(raw_item, dict):
            continue
        total += 1
        completed = raw_item.get("completed") is True
        if completed:
            done += 1
            continue
        if next_text is None:
            text = raw_item.get("text")
            next_text = str(text) if text is not None else None

    return _TodoSummary(done=done, total=total, next_text=next_text)


def _todo_title(summary: _TodoSummary) -> str:
    if summary.total <= 0:
        return "todo"
    if summary.next_text:
        return f"todo {summary.done}/{summary.total}: {summary.next_text}"
    return f"todo {summary.done}/{summary.total}: done"


def _translate_item_event(
    phase: ActionPhase, item: codex_schema.ThreadItem, *, factory: EventFactory
) -> list[TakopiEvent]:
    match item:
        case codex_schema.AgentMessageItem():
            return []
        case codex_schema.ErrorItem(id=action_id, message=message):
            if phase != "completed":
                return []
            return [
                factory.action_completed(
                    action_id=action_id,
                    kind="warning",
                    title=message,
                    detail={"message": message},
                    ok=False,
                    message=message,
                    level="warning",
                ),
            ]
        case codex_schema.CommandExecutionItem(
            id=action_id,
            command=command,
            exit_code=exit_code,
            status=status,
        ):
            title = relativize_command(command)
            if phase in {"started", "updated"}:
                return [
                    factory.action(
                        phase=phase,
                        action_id=action_id,
                        kind="command",
                        title=title,
                    )
                ]
            if phase == "completed":
                ok = status == "completed"
                if isinstance(exit_code, int):
                    ok = ok and exit_code == 0
                detail = {"exit_code": exit_code, "status": status}
                return [
                    factory.action_completed(
                        action_id=action_id,
                        kind="command",
                        title=title,
                        detail=detail,
                        ok=ok,
                    ),
                ]
        case codex_schema.McpToolCallItem(
            id=action_id,
            server=server,
            tool=tool,
            arguments=arguments,
            status=status,
            result=result,
            error=error,
        ):
            title = _short_tool_name(server, tool)
            detail: dict[str, Any] = {
                "server": server,
                "tool": tool,
                "status": status,
                "arguments": arguments,
            }

            if phase in {"started", "updated"}:
                return [
                    factory.action(
                        phase=phase,
                        action_id=action_id,
                        kind="tool",
                        title=title,
                        detail=detail,
                    )
                ]
            if phase == "completed":
                ok = status == "completed" and error is None
                if error is not None:
                    detail["error_message"] = str(error.message)
                result_summary = _summarize_tool_result(result)
                if result_summary is not None:
                    detail["result_summary"] = result_summary
                return [
                    factory.action_completed(
                        action_id=action_id,
                        kind="tool",
                        title=title,
                        detail=detail,
                        ok=ok,
                    ),
                ]
        case codex_schema.WebSearchItem(id=action_id, query=query):
            detail = {"query": query}
            if phase in {"started", "updated"}:
                return [
                    factory.action(
                        phase=phase,
                        action_id=action_id,
                        kind="web_search",
                        title=query,
                        detail=detail,
                    )
                ]
            if phase == "completed":
                return [
                    factory.action_completed(
                        action_id=action_id,
                        kind="web_search",
                        title=query,
                        detail=detail,
                        ok=True,
                    )
                ]
        case codex_schema.FileChangeItem(id=action_id, changes=changes, status=status):
            if phase != "completed":
                return []
            title = _format_change_summary(changes)
            normalized_changes = _normalize_change_list(changes)
            detail = {
                "changes": normalized_changes,
                "status": status,
                "error": None,
            }
            ok = status == "completed"
            return [
                factory.action_completed(
                    action_id=action_id,
                    kind="file_change",
                    title=title,
                    detail=detail,
                    ok=ok,
                )
            ]
        case codex_schema.TodoListItem(id=action_id, items=items):
            summary = _summarize_todo_list(items)
            title = _todo_title(summary)
            detail = {"done": summary.done, "total": summary.total}
            if phase in {"started", "updated"}:
                return [
                    factory.action(
                        phase=phase,
                        action_id=action_id,
                        kind="note",
                        title=title,
                        detail=detail,
                    )
                ]
            if phase == "completed":
                return [
                    factory.action_completed(
                        action_id=action_id,
                        kind="note",
                        title=title,
                        detail=detail,
                        ok=True,
                    )
                ]
        case codex_schema.ReasoningItem(id=action_id, text=text):
            if phase in {"started", "updated"}:
                return [
                    factory.action(
                        phase=phase,
                        action_id=action_id,
                        kind="note",
                        title=text,
                    )
                ]
            if phase == "completed":
                return [
                    factory.action_completed(
                        action_id=action_id,
                        kind="note",
                        title=text,
                        ok=True,
                    )
                ]
    return []


def translate_codex_event(
    event: codex_schema.ThreadEvent,
    *,
    title: str,
    factory: EventFactory,
) -> list[TakopiEvent]:
    match event:
        case codex_schema.ThreadStarted(thread_id=thread_id):
            token = ResumeToken(engine=ENGINE, value=thread_id)
            return [factory.started(token, title=title)]
        case codex_schema.ItemStarted(item=item):
            return _translate_item_event("started", item, factory=factory)
        case codex_schema.ItemUpdated(item=item):
            return _translate_item_event("updated", item, factory=factory)
        case codex_schema.ItemCompleted(item=item):
            return _translate_item_event("completed", item, factory=factory)
        case _:
            return []


@dataclass(slots=True)
class CodexRunState:
    factory: EventFactory
    note_seq: int = 0
    final_answer: str | None = None
    turn_index: int = 0


class CodexRunner(ResumeTokenMixin, JsonlSubprocessRunner):
    engine: EngineId = ENGINE
    resume_re = _RESUME_RE
    logger = logger
    model: str | None = None

    def __init__(
        self,
        *,
        codex_cmd: str,
        extra_args: list[str],
        title: str = "Codex",
    ) -> None:
        self.codex_cmd = codex_cmd
        self.extra_args = extra_args
        self.session_title = title

    def command(self) -> str:
        return self.codex_cmd

    def build_args(
        self,
        prompt: str,
        resume: ResumeToken | None,
        *,
        state: Any,
    ) -> list[str]:
        run_options = get_run_options()
        args = [*self.extra_args]
        if run_options is not None:
            if run_options.model:
                args.extend(["--model", str(run_options.model)])
            if run_options.reasoning:
                args.extend(
                    [
                        "-c",
                        f"model_reasoning_effort={run_options.reasoning}",
                    ]
                )
        args.extend(
            [
                "exec",
                "--json",
                "--skip-git-repo-check",
                "--color=never",
            ]
        )
        if resume:
            args.extend(["resume", resume.value, "-"])
        else:
            args.append("-")
        return args

    def stdin_payload(
        self,
        prompt: str,
        resume: ResumeToken | None,
        *,
        state: Any,
    ) -> bytes | None:
        run_options = get_run_options()
        # Apply system prompt as prefix only on first run (when resume is None)
        if resume is None and run_options is not None and run_options.system:
            prompt = f"{run_options.system}\n\n---\n\n{prompt}"
        return prompt.encode()

    def new_state(self, prompt: str, resume: ResumeToken | None) -> CodexRunState:
        return CodexRunState(factory=EventFactory(ENGINE))

    def start_run(
        self,
        prompt: str,
        resume: ResumeToken | None,
        *,
        state: CodexRunState,
    ) -> None:
        pass

    def decode_jsonl(self, *, line: bytes) -> codex_schema.ThreadEvent:
        return codex_schema.decode_event(line)

    def decode_error_events(
        self,
        *,
        raw: str,
        line: str,
        error: Exception,
        state: CodexRunState,
    ) -> list[TakopiEvent]:
        if isinstance(error, msgspec.DecodeError):
            self.get_logger().warning(
                "jsonl.msgspec.invalid",
                tag=self.tag(),
                error=str(error),
                error_type=error.__class__.__name__,
            )
            return []
        return super().decode_error_events(
            raw=raw,
            line=line,
            error=error,
            state=state,
        )

    def pipes_error_message(self) -> str:
        return "codex exec failed to open subprocess pipes"

    def translate(
        self,
        data: codex_schema.ThreadEvent,
        *,
        state: CodexRunState,
        resume: ResumeToken | None,
        found_session: ResumeToken | None,
    ) -> list[TakopiEvent]:
        factory = state.factory
        match data:
            case codex_schema.StreamError(message=message):
                reconnect = _parse_reconnect_message(message)
                if reconnect is not None:
                    attempt, max_attempts = reconnect
                    phase: ActionPhase = "started" if attempt <= 1 else "updated"
                    return [
                        factory.action(
                            phase=phase,
                            action_id="codex.reconnect",
                            kind="note",
                            title=message,
                            detail={"attempt": attempt, "max": max_attempts},
                            level="info",
                        )
                    ]
                return [self.note_event(message, state=state, ok=False)]
            case codex_schema.TurnFailed(error=error):
                resume_for_completed = found_session or resume
                return [
                    factory.completed_error(
                        error=error.message,
                        answer=state.final_answer or "",
                        resume=resume_for_completed,
                    )
                ]
            case codex_schema.TurnStarted():
                action_id = f"turn_{state.turn_index}"
                state.turn_index += 1
                return [
                    factory.action_started(
                        action_id=action_id,
                        kind="turn",
                        title="turn started",
                    )
                ]
            case codex_schema.TurnCompleted(usage=usage):
                resume_for_completed = found_session or resume
                return [
                    factory.completed_ok(
                        answer=state.final_answer or "",
                        resume=resume_for_completed,
                        usage=msgspec.to_builtins(usage),
                    )
                ]
            case codex_schema.ItemCompleted(
                item=codex_schema.AgentMessageItem(text=text)
            ):
                if state.final_answer is None:
                    state.final_answer = text
                else:
                    logger.debug("codex.multiple_agent_messages")
                    state.final_answer = text
            case _:
                pass

        return translate_codex_event(
            data,
            title=self.session_title,
            factory=factory,
        )

    def process_error_events(
        self,
        rc: int,
        *,
        resume: ResumeToken | None,
        found_session: ResumeToken | None,
        state: CodexRunState,
    ) -> list[TakopiEvent]:
        message = f"codex exec failed (rc={rc})."
        resume_for_completed = found_session or resume
        return [
            self.note_event(
                message,
                state=state,
                ok=False,
            ),
            state.factory.completed_error(
                error=message,
                answer=state.final_answer or "",
                resume=resume_for_completed,
            ),
        ]

    def stream_end_events(
        self,
        *,
        resume: ResumeToken | None,
        found_session: ResumeToken | None,
        state: CodexRunState,
    ) -> list[TakopiEvent]:
        if not found_session:
            message = "codex exec finished but no session_id/thread_id was captured"
            resume_for_completed = resume
            return [
                state.factory.completed_error(
                    error=message,
                    answer=state.final_answer or "",
                    resume=resume_for_completed,
                )
            ]
        logger.info("codex.session.completed", resume=found_session.value)
        return [
            state.factory.completed_ok(
                answer=state.final_answer or "",
                resume=found_session,
            )
        ]


def build_runner(config: EngineConfig, config_path: Path) -> Runner:
    codex_cmd = "codex"

    extra_args_value = config.get("extra_args")
    if extra_args_value is None:
        extra_args = ["-c", "notify=[]"]
    elif isinstance(extra_args_value, list) and all(
        isinstance(item, str) for item in extra_args_value
    ):
        extra_args = list(extra_args_value)
    else:
        raise ConfigError(
            f"Invalid `codex.extra_args` in {config_path}; expected a list of strings."
        )

    exec_only_flag = find_exec_only_flag(extra_args)
    if exec_only_flag:
        raise ConfigError(
            f"Invalid `codex.extra_args` in {config_path}; exec-only flag "
            f"{exec_only_flag!r} is managed by Takopi."
        )

    title = "Codex"
    profile_value = config.get("profile")
    if profile_value:
        if not isinstance(profile_value, str):
            raise ConfigError(
                f"Invalid `codex.profile` in {config_path}; expected a string."
            )
        extra_args.extend(["--profile", profile_value])
        title = profile_value

    return CodexRunner(codex_cmd=codex_cmd, extra_args=extra_args, title=title)


BACKEND = EngineBackend(
    id="codex",
    build_runner=build_runner,
    install_cmd="npm install -g @openai/codex",
)



from pathlib import Path

from yee88.config import ProjectConfig, ProjectsConfig
from yee88.ids import RESERVED_CHAT_COMMANDS
from yee88.router import AutoRouter, RunnerEntry
from yee88.runners.mock import Return, ScriptRunner
from yee88.telegram.trigger_mode import should_trigger_run
from yee88.telegram.types import TelegramIncomingMessage
from yee88.transport_runtime import TransportRuntime


def _runtime() -> TransportRuntime:
    runner = ScriptRunner([Return(answer="ok")], engine="codex")
    router = AutoRouter(
        entries=[RunnerEntry(engine=runner.engine, runner=runner)],
        default_engine=runner.engine,
    )
    projects = ProjectsConfig(
        projects={
            "proj": ProjectConfig(
                alias="proj",
                path=Path("."),
                worktrees_dir=Path(".worktrees"),
            )
        },
        default_project=None,
    )
    return TransportRuntime(router=router, projects=projects)


def _msg(text: str, **kwargs) -> TelegramIncomingMessage:
    return TelegramIncomingMessage(
        transport="telegram",
        chat_id=1,
        message_id=1,
        text=text,
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=1,
        **kwargs,
    )


def test_should_trigger_run_mentions() -> None:
    runtime = _runtime()
    msg = _msg("hello @bot")
    assert should_trigger_run(
        msg,
        bot_username="bot",
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )


def test_should_trigger_run_engine_and_project() -> None:
    runtime = _runtime()
    assert should_trigger_run(
        _msg("/codex hello"),
        bot_username=None,
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )
    assert should_trigger_run(
        _msg("/proj hello"),
        bot_username=None,
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )


def test_should_trigger_run_reply_to_bot() -> None:
    runtime = _runtime()
    msg = _msg("hello", reply_to_is_bot=True)
    assert should_trigger_run(
        msg,
        bot_username=None,
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )


def test_should_trigger_run_ignores_implicit_topic_reply_to_root() -> None:
    runtime = _runtime()
    msg = TelegramIncomingMessage(
        transport="telegram",
        chat_id=1,
        message_id=187,
        text="hello",
        reply_to_message_id=163,
        reply_to_text=None,
        reply_to_is_bot=True,
        reply_to_username="TakopiBot",
        sender_id=1,
        thread_id=163,
        is_topic_message=True,
        chat_type="supergroup",
        is_forum=True,
    )
    assert not should_trigger_run(
        msg,
        bot_username=None,
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )


def test_should_trigger_run_known_commands() -> None:
    runtime = _runtime()
    assert should_trigger_run(
        _msg("/agent"),
        bot_username=None,
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )
    assert should_trigger_run(
        _msg("/ping"),
        bot_username=None,
        runtime=runtime,
        command_ids={"ping"},
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )


def test_should_trigger_run_ignores_unknown_commands() -> None:
    runtime = _runtime()
    assert not should_trigger_run(
        _msg("/wat"),
        bot_username=None,
        runtime=runtime,
        command_ids=set(),
        reserved_chat_commands=set(RESERVED_CHAT_COMMANDS),
    )



from dataclasses import replace
from pathlib import Path

import pytest

from yee88.telegram.api_models import ChatMember
from yee88.telegram.commands.agent import _handle_agent_command
from yee88.telegram.commands.trigger import _handle_trigger_command
from yee88.telegram.chat_prefs import ChatPrefsStore
from yee88.telegram.topic_state import TopicStateStore
from yee88.telegram.types import TelegramIncomingMessage
from yee88.settings import TelegramTopicsSettings
from tests.telegram_fakes import FakeBot, FakeTransport, make_cfg


def _msg(
    text: str,
    *,
    chat_id: int = 123,
    message_id: int = 10,
    sender_id: int | None = 42,
    chat_type: str | None = "private",
    thread_id: int | None = None,
) -> TelegramIncomingMessage:
    return TelegramIncomingMessage(
        transport="telegram",
        chat_id=chat_id,
        message_id=message_id,
        text=text,
        reply_to_message_id=None,
        reply_to_text=None,
        sender_id=sender_id,
        chat_type=chat_type,
        thread_id=thread_id,
    )


def _last_text(transport: FakeTransport) -> str:
    assert transport.send_calls
    return transport.send_calls[-1]["message"].text


class _MemberBot(FakeBot):
    async def get_chat_member(self, chat_id: int, user_id: int) -> ChatMember | None:
        _ = chat_id, user_id
        return ChatMember(status="member", can_manage_topics=False)


@pytest.mark.anyio
async def test_agent_show_private_defaults() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    msg = _msg("/agent", chat_type="private")

    await _handle_agent_command(
        cfg,
        msg,
        args_text="",
        ambient_context=None,
        topic_store=None,
        chat_prefs=None,
    )

    text = _last_text(transport)
    assert "engine: codex" in text
    assert "available: codex" in text


@pytest.mark.anyio
async def test_agent_set_clear_group_admin(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    prefs = ChatPrefsStore(tmp_path / "prefs.json")
    msg = _msg("/agent set codex", chat_type="supergroup")

    await _handle_agent_command(
        cfg,
        msg,
        args_text="set codex",
        ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )

    assert await prefs.get_default_engine(msg.chat_id) == "codex"
    assert "chat default engine set" in _last_text(transport)

    await _handle_agent_command(
        cfg,
        msg,
        args_text="clear",
        ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )

    assert await prefs.get_default_engine(msg.chat_id) is None
    assert "chat default engine cleared" in _last_text(transport)


@pytest.mark.anyio
async def test_agent_set_denied_for_non_admin(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = replace(make_cfg(transport), bot=_MemberBot())
    prefs = ChatPrefsStore(tmp_path / "prefs.json")
    msg = _msg("/agent set codex", chat_type="supergroup")

    await _handle_agent_command(
        cfg,
        msg,
        args_text="set codex",
        ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )

    assert await prefs.get_default_engine(msg.chat_id) is None
    assert "restricted to group admins" in _last_text(transport)


@pytest.mark.anyio
async def test_agent_set_invalid_engine(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    msg = _msg("/agent set nope", chat_type="private")

    await _handle_agent_command(
        cfg,
        msg,
        args_text="set nope",
        ambient_context=None,
        topic_store=None,
        chat_prefs=ChatPrefsStore(tmp_path / "prefs.json"),
    )

    text = _last_text(transport)
    assert "unknown engine" in text
    assert "available engines" in text


@pytest.mark.anyio
@pytest.mark.parametrize(
    ("topic_mode", "chat_mode", "expected_source", "expected_trigger"),
    [
        ("mentions", None, "topic override", "mentions"),
        (None, "mentions", "chat default", "mentions"),
        (None, None, "default", "all"),
    ],
)
async def test_trigger_show_sources(
    tmp_path: Path,
    topic_mode: str | None,
    chat_mode: str | None,
    expected_source: str,
    expected_trigger: str,
) -> None:
    transport = FakeTransport()
    cfg = replace(
        make_cfg(transport),
        topics=TelegramTopicsSettings(enabled=True, scope="all"),
    )
    topic_store = TopicStateStore(tmp_path / "topics.json")
    chat_prefs = ChatPrefsStore(tmp_path / "prefs.json")
    msg = _msg("/trigger", chat_type="supergroup", thread_id=7)

    if topic_mode is not None:
        await topic_store.set_trigger_mode(msg.chat_id, msg.thread_id or 0, topic_mode)
    if chat_mode is not None:
        await chat_prefs.set_trigger_mode(msg.chat_id, chat_mode)

    await _handle_trigger_command(
        cfg,
        msg,
        args_text="",
        _ambient_context=None,
        topic_store=topic_store,
        chat_prefs=chat_prefs,
    )

    text = _last_text(transport)
    assert f"trigger: {expected_trigger} ({expected_source})" in text
    assert "available: all, mentions" in text


@pytest.mark.anyio
async def test_trigger_set_clear_permissions(tmp_path: Path) -> None:
    transport = FakeTransport()
    prefs = ChatPrefsStore(tmp_path / "prefs.json")
    msg = _msg("/trigger mentions", chat_type="supergroup")

    denied_cfg = replace(make_cfg(transport), bot=_MemberBot())
    await _handle_trigger_command(
        denied_cfg,
        msg,
        args_text="mentions",
        _ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )
    assert await prefs.get_trigger_mode(msg.chat_id) is None
    assert "restricted to group admins" in _last_text(transport)

    transport = FakeTransport()
    allow_cfg = make_cfg(transport)
    await _handle_trigger_command(
        allow_cfg,
        msg,
        args_text="mentions",
        _ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )
    assert await prefs.get_trigger_mode(msg.chat_id) == "mentions"
    assert "chat trigger mode set" in _last_text(transport)

    await _handle_trigger_command(
        allow_cfg,
        msg,
        args_text="clear",
        _ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )
    assert await prefs.get_trigger_mode(msg.chat_id) is None
    assert "chat trigger mode reset" in _last_text(transport)


@pytest.mark.anyio
async def test_trigger_missing_sender_denied(tmp_path: Path) -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    prefs = ChatPrefsStore(tmp_path / "prefs.json")
    msg = _msg("/trigger all", chat_type="supergroup", sender_id=None)

    await _handle_trigger_command(
        cfg,
        msg,
        args_text="all",
        _ambient_context=None,
        topic_store=None,
        chat_prefs=prefs,
    )

    assert await prefs.get_trigger_mode(msg.chat_id) is None
    assert "cannot verify sender" in _last_text(transport)


@pytest.mark.anyio
async def test_trigger_topic_unavailable() -> None:
    transport = FakeTransport()
    cfg = replace(
        make_cfg(transport),
        topics=TelegramTopicsSettings(enabled=True, scope="all"),
    )
    msg = _msg("/trigger mentions", chat_type="supergroup", thread_id=3)

    await _handle_trigger_command(
        cfg,
        msg,
        args_text="mentions",
        _ambient_context=None,
        topic_store=None,
        chat_prefs=None,
    )

    assert "topic trigger settings are unavailable" in _last_text(transport)


@pytest.mark.anyio
async def test_trigger_chat_prefs_unavailable() -> None:
    transport = FakeTransport()
    cfg = make_cfg(transport)
    msg = _msg("/trigger mentions", chat_type="supergroup")

    await _handle_trigger_command(
        cfg,
        msg,
        args_text="mentions",
        _ambient_context=None,
        topic_store=None,
        chat_prefs=None,
    )

    assert "chat trigger settings are unavailable" in _last_text(transport)



from __future__ import annotations

# Headless JSONL schema derived from tag rust-v0.77.0 (git 112f40e91c12af0f7146d7e03f20283516a8af0b).

from typing import Any, Literal

import msgspec

type CommandExecutionStatus = Literal[
    "in_progress",
    "completed",
    "failed",
    "declined",
]
type PatchApplyStatus = Literal[
    "in_progress",
    "completed",
    "failed",
]
type PatchChangeKind = Literal[
    "add",
    "delete",
    "update",
]
type McpToolCallStatus = Literal[
    "in_progress",
    "completed",
    "failed",
]


class Usage(msgspec.Struct, kw_only=True):
    input_tokens: int
    cached_input_tokens: int
    output_tokens: int


class ThreadError(msgspec.Struct, kw_only=True):
    message: str


class ThreadStarted(msgspec.Struct, tag="thread.started", kw_only=True):
    thread_id: str


class TurnStarted(msgspec.Struct, tag="turn.started", kw_only=True):
    pass


class TurnCompleted(msgspec.Struct, tag="turn.completed", kw_only=True):
    usage: Usage


class TurnFailed(msgspec.Struct, tag="turn.failed", kw_only=True):
    error: ThreadError


class StreamError(msgspec.Struct, tag="error", kw_only=True):
    message: str


class AgentMessageItem(msgspec.Struct, tag="agent_message", kw_only=True):
    id: str
    text: str


class ReasoningItem(msgspec.Struct, tag="reasoning", kw_only=True):
    id: str
    text: str


class CommandExecutionItem(msgspec.Struct, tag="command_execution", kw_only=True):
    id: str
    command: str
    aggregated_output: str
    exit_code: int | None
    status: CommandExecutionStatus


class FileUpdateChange(msgspec.Struct, kw_only=True):
    path: str
    kind: PatchChangeKind


class FileChangeItem(msgspec.Struct, tag="file_change", kw_only=True):
    id: str
    changes: list[FileUpdateChange]
    status: PatchApplyStatus


class McpToolCallItemResult(msgspec.Struct, kw_only=True):
    content: list[dict[str, Any]]
    structured_content: Any


class McpToolCallItemError(msgspec.Struct, kw_only=True):
    message: str


class McpToolCallItem(msgspec.Struct, tag="mcp_tool_call", kw_only=True):
    id: str
    server: str
    tool: str
    arguments: Any
    result: McpToolCallItemResult | None
    error: McpToolCallItemError | None
    status: McpToolCallStatus


class WebSearchItem(msgspec.Struct, tag="web_search", kw_only=True):
    id: str
    query: str


class ErrorItem(msgspec.Struct, tag="error", kw_only=True):
    id: str
    message: str


class TodoItem(msgspec.Struct, kw_only=True):
    text: str
    completed: bool


class TodoListItem(msgspec.Struct, tag="todo_list", kw_only=True):
    id: str
    items: list[TodoItem]


type ThreadItem = (
    AgentMessageItem
    | ReasoningItem
    | CommandExecutionItem
    | FileChangeItem
    | McpToolCallItem
    | WebSearchItem
    | TodoListItem
    | ErrorItem
)


class ItemStarted(msgspec.Struct, tag="item.started", kw_only=True):
    item: ThreadItem


class ItemUpdated(msgspec.Struct, tag="item.updated", kw_only=True):
    item: ThreadItem


class ItemCompleted(msgspec.Struct, tag="item.completed", kw_only=True):
    item: ThreadItem


type ThreadEvent = (
    ThreadStarted
    | TurnStarted
    | TurnCompleted
    | TurnFailed
    | ItemStarted
    | ItemUpdated
    | ItemCompleted
    | StreamError
)

_DECODER = msgspec.json.Decoder(ThreadEvent)


def decode_event(data: bytes | str) -> ThreadEvent:
    return _DECODER.decode(data)



from __future__ import annotations

import time
from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field

import anyio

from .context import RunContext
from .logging import bind_run_context, get_logger
from .model import CompletedEvent, ResumeToken, StartedEvent, TakopiEvent
from .runners.run_options import get_run_options
from .presenter import Presenter
from .markdown import render_event_cli
from .runner import Runner
from .progress import ProgressTracker
from .transport import (
    ChannelId,
    MessageId,
    MessageRef,
    RenderedMessage,
    SendOptions,
    ThreadId,
    Transport,
)

logger = get_logger(__name__)


def _effective_model(runner: Runner) -> str | None:
    run_options = get_run_options()
    if run_options is not None and run_options.model:
        return run_options.model
    return runner.model


def _log_runner_event(evt: TakopiEvent) -> None:
    for line in render_event_cli(evt):
        logger.debug(
            "runner.event.cli",
            line=line,
            event_type=getattr(evt, "type", None),
            engine=getattr(evt, "engine", None),
        )


def _strip_resume_lines(text: str, *, is_resume_line: Callable[[str], bool]) -> str:
    prompt = "\n".join(
        line for line in text.splitlines() if not is_resume_line(line)
    ).strip()
    return prompt or "continue"


def _flatten_exception_group(error: BaseException) -> list[BaseException]:
    if isinstance(error, BaseExceptionGroup):
        flattened: list[BaseException] = []
        for exc in error.exceptions:
            flattened.extend(_flatten_exception_group(exc))
        return flattened
    return [error]


def _format_error(error: Exception) -> str:
    cancel_exc = anyio.get_cancelled_exc_class()
    flattened = [
        exc
        for exc in _flatten_exception_group(error)
        if not isinstance(exc, cancel_exc)
    ]
    if len(flattened) == 1:
        return str(flattened[0]) or flattened[0].__class__.__name__
    if not flattened:
        return str(error) or error.__class__.__name__
    messages = [str(exc) for exc in flattened if str(exc)]
    if not messages:
        return str(error) or error.__class__.__name__
    if len(messages) == 1:
        return messages[0]
    return "\n".join(messages)


@dataclass(frozen=True, slots=True)
class IncomingMessage:
    channel_id: ChannelId
    message_id: MessageId
    text: str
    reply_to: MessageRef | None = None
    thread_id: ThreadId | None = None


@dataclass(frozen=True, slots=True)
class ExecBridgeConfig:
    transport: Transport
    presenter: Presenter
    final_notify: bool


@dataclass(slots=True)
class RunningTask:
    resume: ResumeToken | None = None
    resume_ready: anyio.Event = field(default_factory=anyio.Event)
    cancel_requested: anyio.Event = field(default_factory=anyio.Event)
    done: anyio.Event = field(default_factory=anyio.Event)
    context: RunContext | None = None


RunningTasks = dict[MessageRef, RunningTask]


async def _send_or_edit_message(
    transport: Transport,
    *,
    channel_id: ChannelId,
    message: RenderedMessage,
    edit_ref: MessageRef | None = None,
    reply_to: MessageRef | None = None,
    notify: bool = True,
    replace_ref: MessageRef | None = None,
    thread_id: ThreadId | None = None,
) -> tuple[MessageRef | None, bool]:
    msg = message
    followups = message.extra.get("followups")
    if followups:
        extra = dict(message.extra)
        if reply_to is not None:
            extra.setdefault("followup_reply_to_message_id", reply_to.message_id)
        if thread_id is not None:
            extra.setdefault("followup_thread_id", thread_id)
        extra.setdefault("followup_notify", notify)
        msg = RenderedMessage(text=message.text, extra=extra)
    if edit_ref is not None:
        logger.debug(
            "transport.edit_message",
            channel_id=edit_ref.channel_id,
            message_id=edit_ref.message_id,
            rendered=msg.text,
        )
        edited = await transport.edit(ref=edit_ref, message=msg)
        if edited is not None:
            return edited, True

    logger.debug(
        "transport.send_message",
        channel_id=channel_id,
        reply_to_message_id=reply_to.message_id if reply_to else None,
        rendered=msg.text,
    )
    sent = await transport.send(
        channel_id=channel_id,
        message=msg,
        options=SendOptions(
            reply_to=reply_to,
            notify=notify,
            replace=replace_ref,
            thread_id=thread_id,
        ),
    )
    return sent, False


class ProgressEdits:
    def __init__(
        self,
        *,
        transport: Transport,
        presenter: Presenter,
        channel_id: ChannelId,
        progress_ref: MessageRef | None,
        tracker: ProgressTracker,
        started_at: float,
        clock: Callable[[], float],
        last_rendered: RenderedMessage | None,
        resume_formatter: Callable[[ResumeToken], str] | None = None,
        label: str = "working",
        context_line: str | None = None,
        model: str | None = None,
    ) -> None:
        self.transport = transport
        self.presenter = presenter
        self.channel_id = channel_id
        self.progress_ref = progress_ref
        self.tracker = tracker
        self.started_at = started_at
        self.clock = clock
        self.last_rendered = last_rendered
        self.resume_formatter = resume_formatter
        self.label = label
        self.context_line = context_line
        self.model = model
        self.event_seq = 0
        self.rendered_seq = 0
        self.signal_send, self.signal_recv = anyio.create_memory_object_stream(1)

    async def run(self) -> None:
        if self.progress_ref is None:
            return
        while True:
            while self.rendered_seq == self.event_seq:
                try:
                    await self.signal_recv.receive()
                except anyio.EndOfStream:
                    return

            seq_at_render = self.event_seq
            now = self.clock()
            state = self.tracker.snapshot(
                resume_formatter=self.resume_formatter,
                context_line=self.context_line,
                model=self.model,
            )
            rendered = self.presenter.render_progress(
                state, elapsed_s=now - self.started_at, label=self.label
            )
            if rendered != self.last_rendered:
                logger.debug(
                    "transport.edit_message",
                    channel_id=self.channel_id,
                    message_id=self.progress_ref.message_id,
                    rendered=rendered.text,
                )
                edited = await self.transport.edit(
                    ref=self.progress_ref,
                    message=rendered,
                    wait=False,
                )
                if edited is not None:
                    self.last_rendered = rendered

            self.rendered_seq = seq_at_render

    async def on_event(self, evt: TakopiEvent) -> None:
        if not self.tracker.note_event(evt):
            return
        if self.progress_ref is None:
            return
        self.event_seq += 1
        try:
            self.signal_send.send_nowait(None)
        except anyio.WouldBlock:
            pass
        except (anyio.BrokenResourceError, anyio.ClosedResourceError):
            pass


@dataclass(frozen=True, slots=True)
class ProgressMessageState:
    ref: MessageRef | None
    last_rendered: RenderedMessage | None


async def send_initial_progress(
    cfg: ExecBridgeConfig,
    *,
    channel_id: ChannelId,
    reply_to: MessageRef,
    label: str,
    tracker: ProgressTracker,
    progress_ref: MessageRef | None = None,
    resume_formatter: Callable[[ResumeToken], str] | None = None,
    context_line: str | None = None,
    thread_id: ThreadId | None = None,
    model: str | None = None,
) -> ProgressMessageState:
    last_rendered: RenderedMessage | None = None

    state = tracker.snapshot(
        resume_formatter=resume_formatter,
        context_line=context_line,
        model=model,
    )
    initial_rendered = cfg.presenter.render_progress(
        state,
        elapsed_s=0.0,
        label=label,
    )
    sent_ref, _ = await _send_or_edit_message(
        cfg.transport,
        channel_id=channel_id,
        message=initial_rendered,
        edit_ref=progress_ref,
        reply_to=reply_to,
        notify=False,
        replace_ref=progress_ref,
        thread_id=thread_id,
    )
    if sent_ref is not None:
        last_rendered = initial_rendered
        logger.debug(
            "progress.sent",
            channel_id=sent_ref.channel_id,
            message_id=sent_ref.message_id,
        )

    return ProgressMessageState(
        ref=sent_ref,
        last_rendered=last_rendered,
    )


@dataclass(slots=True)
class RunOutcome:
    cancelled: bool = False
    completed: CompletedEvent | None = None
    resume: ResumeToken | None = None


async def run_runner_with_cancel(
    runner: Runner,
    *,
    prompt: str,
    resume_token: ResumeToken | None,
    edits: ProgressEdits,
    running_task: RunningTask | None,
    on_thread_known: Callable[[ResumeToken, anyio.Event], Awaitable[None]] | None,
) -> RunOutcome:
    outcome = RunOutcome()
    async with anyio.create_task_group() as tg:

        async def run_runner() -> None:
            try:
                async for evt in runner.run(prompt, resume_token):
                    _log_runner_event(evt)
                    if isinstance(evt, StartedEvent):
                        outcome.resume = evt.resume
                        bind_run_context(resume=evt.resume.value)
                        if running_task is not None and running_task.resume is None:
                            running_task.resume = evt.resume
                            try:
                                if on_thread_known is not None:
                                    await on_thread_known(evt.resume, running_task.done)
                            finally:
                                running_task.resume_ready.set()
                    elif isinstance(evt, CompletedEvent):
                        outcome.resume = evt.resume or outcome.resume
                        outcome.completed = evt
                    await edits.on_event(evt)
            finally:
                tg.cancel_scope.cancel()

        async def wait_cancel(task: RunningTask) -> None:
            await task.cancel_requested.wait()
            outcome.cancelled = True
            tg.cancel_scope.cancel()

        tg.start_soon(run_runner)
        if running_task is not None:
            tg.start_soon(wait_cancel, running_task)

    return outcome


def sync_resume_token(
    tracker: ProgressTracker, resume: ResumeToken | None
) -> ResumeToken | None:
    resume = resume or tracker.resume
    tracker.set_resume(resume)
    return resume


async def send_result_message(
    cfg: ExecBridgeConfig,
    *,
    channel_id: ChannelId,
    reply_to: MessageRef,
    progress_ref: MessageRef | None,
    message: RenderedMessage,
    notify: bool,
    edit_ref: MessageRef | None,
    replace_ref: MessageRef | None = None,
    delete_tag: str = "final",
    thread_id: ThreadId | None = None,
) -> None:
    final_msg, edited = await _send_or_edit_message(
        cfg.transport,
        channel_id=channel_id,
        message=message,
        edit_ref=edit_ref,
        reply_to=reply_to,
        notify=notify,
        replace_ref=replace_ref,
        thread_id=thread_id,
    )
    if final_msg is None:
        return
    if (
        progress_ref is not None
        and (edit_ref is None or not edited)
        and replace_ref is None
    ):
        logger.debug(
            "transport.delete_message",
            channel_id=progress_ref.channel_id,
            message_id=progress_ref.message_id,
            tag=delete_tag,
        )
        await cfg.transport.delete(ref=progress_ref)


async def handle_message(
    cfg: ExecBridgeConfig,
    *,
    runner: Runner,
    incoming: IncomingMessage,
    resume_token: ResumeToken | None,
    context: RunContext | None = None,
    context_line: str | None = None,
    strip_resume_line: Callable[[str], bool] | None = None,
    running_tasks: RunningTasks | None = None,
    on_thread_known: Callable[[ResumeToken, anyio.Event], Awaitable[None]]
    | None = None,
    progress_ref: MessageRef | None = None,
    clock: Callable[[], float] = time.monotonic,
) -> None:
    logger.info(
        "handle.incoming",
        channel_id=incoming.channel_id,
        user_msg_id=incoming.message_id,
        resume=resume_token.value if resume_token else None,
        text=incoming.text,
    )
    started_at = clock()
    is_resume_line = runner.is_resume_line
    resume_strip = strip_resume_line or is_resume_line
    runner_text = _strip_resume_lines(incoming.text, is_resume_line=resume_strip)

    progress_tracker = ProgressTracker(engine=runner.engine)

    user_ref = MessageRef(
        channel_id=incoming.channel_id,
        message_id=incoming.message_id,
    )
    progress_state = await send_initial_progress(
        cfg,
        channel_id=incoming.channel_id,
        reply_to=user_ref,
        label="starting",
        tracker=progress_tracker,
        progress_ref=progress_ref,
        resume_formatter=runner.format_resume,
        context_line=context_line,
        thread_id=incoming.thread_id,
        model=_effective_model(runner),
    )
    progress_ref = progress_state.ref

    edits = ProgressEdits(
        transport=cfg.transport,
        presenter=cfg.presenter,
        channel_id=incoming.channel_id,
        progress_ref=progress_ref,
        tracker=progress_tracker,
        started_at=started_at,
        clock=clock,
        last_rendered=progress_state.last_rendered,
        resume_formatter=runner.format_resume,
        context_line=context_line,
        model=_effective_model(runner),
    )

    running_task: RunningTask | None = None
    if running_tasks is not None and progress_ref is not None:
        running_task = RunningTask(context=context)
        running_tasks[progress_ref] = running_task

    cancel_exc_type = anyio.get_cancelled_exc_class()
    edits_scope = anyio.CancelScope()

    async def run_edits() -> None:
        try:
            with edits_scope:
                await edits.run()
        except cancel_exc_type:
            # Edits are best-effort; cancellation should not bubble into the task group.
            return

    outcome = RunOutcome()
    error: Exception | None = None

    async with anyio.create_task_group() as tg:
        if progress_ref is not None:
            tg.start_soon(run_edits)

        try:
            outcome = await run_runner_with_cancel(
                runner,
                prompt=runner_text,
                resume_token=resume_token,
                edits=edits,
                running_task=running_task,
                on_thread_known=on_thread_known,
            )
        except Exception as exc:
            error = exc
            logger.exception(
                "handle.runner_failed",
                error=str(exc),
                error_type=exc.__class__.__name__,
            )
        finally:
            if running_task is not None and running_tasks is not None:
                running_task.done.set()
                if progress_ref is not None:
                    running_tasks.pop(progress_ref, None)
            if not outcome.cancelled and error is None:
                # Give pending progress edits a chance to flush if they're ready.
                await anyio.sleep(0)
            edits_scope.cancel()

    elapsed = clock() - started_at

    if error is not None:
        sync_resume_token(progress_tracker, outcome.resume)
        err_body = _format_error(error)
        state = progress_tracker.snapshot(
            resume_formatter=runner.format_resume,
            context_line=context_line,
            model=_effective_model(runner),
        )
        final_rendered = cfg.presenter.render_final(
            state,
            elapsed_s=elapsed,
            status="error",
            answer=err_body,
        )
        logger.debug(
            "handle.error.rendered",
            error=err_body,
            rendered=final_rendered.text,
        )
        await send_result_message(
            cfg,
            channel_id=incoming.channel_id,
            reply_to=user_ref,
            progress_ref=progress_ref,
            message=final_rendered,
            notify=False,
            edit_ref=progress_ref,
            replace_ref=progress_ref,
            delete_tag="error",
            thread_id=incoming.thread_id,
        )
        return

    if outcome.cancelled:
        resume = sync_resume_token(progress_tracker, outcome.resume)
        logger.info(
            "handle.cancelled",
            resume=resume.value if resume else None,
            elapsed_s=elapsed,
        )
        state = progress_tracker.snapshot(
            resume_formatter=runner.format_resume,
            context_line=context_line,
            model=_effective_model(runner),
        )
        final_rendered = cfg.presenter.render_progress(
            state,
            elapsed_s=elapsed,
            label="`cancelled`",
        )
        await send_result_message(
            cfg,
            channel_id=incoming.channel_id,
            reply_to=user_ref,
            progress_ref=progress_ref,
            message=final_rendered,
            notify=False,
            edit_ref=progress_ref,
            replace_ref=progress_ref,
            delete_tag="cancel",
            thread_id=incoming.thread_id,
        )
        return

    if outcome.completed is None:
        raise RuntimeError("runner finished without a completed event")

    completed = outcome.completed
    run_ok = completed.ok
    run_error = completed.error

    final_answer = completed.answer
    if run_ok is False and run_error:
        if final_answer.strip():
            final_answer = f"{final_answer}\n\n{run_error}"
        else:
            final_answer = str(run_error)

    status = (
        "error" if run_ok is False else ("done" if final_answer.strip() else "error")
    )
    resume_value = None
    resume_token = completed.resume or outcome.resume
    if resume_token is not None:
        resume_value = resume_token.value
    logger.info(
        "runner.completed",
        ok=run_ok,
        error=run_error,
        answer_len=len(final_answer or ""),
        elapsed_s=round(elapsed, 2),
        action_count=progress_tracker.action_count,
        resume=resume_value,
    )
    sync_resume_token(progress_tracker, completed.resume or outcome.resume)
    state = progress_tracker.snapshot(
        resume_formatter=runner.format_resume,
        context_line=context_line,
        model=_effective_model(runner),
    )
    final_rendered = cfg.presenter.render_final(
        state,
        elapsed_s=elapsed,
        status=status,
        answer=final_answer,
    )
    logger.debug(
        "handle.final.rendered",
        rendered=final_rendered.text,
        status=status,
    )

    can_edit_final = progress_ref is not None
    edit_ref = None if cfg.final_notify or not can_edit_final else progress_ref

    await send_result_message(
        cfg,
        channel_id=incoming.channel_id,
        reply_to=user_ref,
        progress_ref=progress_ref,
        message=final_rendered,
        notify=cfg.final_notify,
        edit_ref=edit_ref,
        replace_ref=progress_ref,
        delete_tag="final",
        thread_id=incoming.thread_id,
    )



from __future__ import annotations

import subprocess
from collections.abc import Sequence
from pathlib import Path


def _run_git(
    args: Sequence[str], *, cwd: Path
) -> subprocess.CompletedProcess[str] | None:
    try:
        return subprocess.run(
            ["git", *args],
            cwd=cwd,
            check=False,
            text=True,
            capture_output=True,
        )
    except FileNotFoundError:
        return None


def git_run(
    args: Sequence[str], *, cwd: Path
) -> subprocess.CompletedProcess[str] | None:
    return _run_git(args, cwd=cwd)


def git_stdout(args: Sequence[str], *, cwd: Path) -> str | None:
    result = _run_git(args, cwd=cwd)
    if result is None or result.returncode != 0:
        return None
    output = result.stdout.strip()
    return output or None


def git_ok(args: Sequence[str], *, cwd: Path) -> bool:
    result = _run_git(args, cwd=cwd)
    return result is not None and result.returncode == 0


def git_is_worktree(path: Path) -> bool:
    top = git_stdout(
        ["rev-parse", "--path-format=absolute", "--show-toplevel"],
        cwd=path,
    )
    if not top:
        return False
    return Path(top).resolve(strict=False) == path.resolve(strict=False)


def resolve_default_base(root: Path) -> str | None:
    origin_head = git_stdout(
        ["symbolic-ref", "-q", "refs/remotes/origin/HEAD"],
        cwd=root,
    )
    if origin_head:
        prefix = "refs/remotes/origin/"
        if origin_head.startswith(prefix):
            name = origin_head[len(prefix) :].strip()
            if name:
                return name

    current = git_stdout(["branch", "--show-current"], cwd=root)
    if current:
        return current

    if git_ok(["show-ref", "--verify", "--quiet", "refs/heads/master"], cwd=root):
        return "master"
    if git_ok(["show-ref", "--verify", "--quiet", "refs/heads/main"], cwd=root):
        return "main"
    return None


def resolve_main_worktree_root(cwd: Path) -> Path | None:
    common_dir = git_stdout(
        ["rev-parse", "--path-format=absolute", "--git-common-dir"],
        cwd=cwd,
    )
    if not common_dir:
        return None
    if git_stdout(["rev-parse", "--is-bare-repository"], cwd=cwd) == "true":
        return cwd
    common_path = Path(common_dir)
    if not common_path.is_absolute():
        common_path = (cwd / common_path).resolve()
    return common_path.parent



from __future__ import annotations

from pathlib import Path

from yee88.utils.paths import (
    relativize_command,
    relativize_path,
    reset_run_base_dir,
    set_run_base_dir,
)


def test_relativize_command_rewrites_cwd_paths(tmp_path: Path) -> None:
    base = tmp_path / "repo"
    base.mkdir()
    command = f'find {base}/tests -type f -name "*.py" | head -20'
    expected = 'find tests -type f -name "*.py" | head -20'
    assert relativize_command(command, base_dir=base) == expected


def test_relativize_command_rewrites_equals_paths(tmp_path: Path) -> None:
    base = tmp_path / "repo"
    base.mkdir()
    command = f'rg -n --files -g "*.py" --path={base}/src'
    expected = 'rg -n --files -g "*.py" --path=src'
    assert relativize_command(command, base_dir=base) == expected


def test_relativize_path_ignores_sibling_prefix(tmp_path: Path) -> None:
    base = tmp_path / "repo"
    base.mkdir()
    value = str(tmp_path / "repo2" / "file.txt")
    assert relativize_path(value, base_dir=base) == value


def test_relativize_path_inside_base(tmp_path: Path) -> None:
    base = tmp_path / "repo"
    base.mkdir()
    value = str(base / "src" / "app.py")
    assert relativize_path(value, base_dir=base) == "src/app.py"


def test_relativize_path_uses_run_base_dir(tmp_path: Path) -> None:
    base = tmp_path / "repo"
    base.mkdir()
    token = set_run_base_dir(base)
    try:
        value = str(base / "src" / "app.py")
        assert relativize_path(value) == "src/app.py"
    finally:
        reset_run_base_dir(token)
